-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.i2c_pkg.ALL;


-- Need to wrap i2c_master with this avs_i2c_master, because the SOPC Component
-- Editor does not support the user defined types from i2c_pkg.

-- Name signals for automatic type and interface recognition:
--
--  <interface type>_<interface name>_<signal type>[_n]
--
-- * Avalon <interface type>:
--   . csi_ = Clock input
--   . avs_ = Avalon-MM slave
--   . coe_ = Conduit
-- * User defined <interface name>:
--   . _system_
--   . _control_
--   . etc.
-- * Avalon <signal type>:
--   . _reset
--   . _clk
--   . _address
--   . etc.

ENTITY avs_i2c_master IS
  GENERIC (
    -- g_i2c_mm
    g_control_adr_w          : NATURAL := c_i2c_control_adr_w;   -- =1, fixed
    g_protocol_adr_w         : NATURAL := c_i2c_protocol_adr_w;
    g_result_adr_w           : NATURAL := c_i2c_result_adr_w;
    -- g_i2c_phy
    g_clk_cnt                : NATURAL := 399;
    g_comma_w                : NATURAL := c_i2c_comma_w_dis
  );
  PORT (
    ---------------------------------------------------------------------------
    -- Avalon Conduit interfaces: coe_*_export
    ---------------------------------------------------------------------------
    -- GENERIC Signal
    coe_gs_sim_export        : IN  STD_LOGIC := '0';

    -- System
    coe_sync_export          : IN  STD_LOGIC := '1';
    
    -- I2C interface
    coe_i2c_scl_export       : INOUT STD_LOGIC;  -- I2C Serial Clock Line
    coe_i2c_sda_export       : INOUT STD_LOGIC;  -- I2C Serial Data Line
    
    ---------------------------------------------------------------------------
    -- Avalon Clock Input interface: csi_*
    ---------------------------------------------------------------------------
    csi_system_reset         : IN  STD_LOGIC;
    csi_system_clk           : IN  STD_LOGIC;
    
    ---------------------------------------------------------------------------
    -- Avalon Memory Mapped Slave interfaces: avs_*
    ---------------------------------------------------------------------------
    -- MM slave I2C control register
    avs_control_address      : IN  STD_LOGIC;  -- g_control_adr_w=1
    avs_control_write        : IN  STD_LOGIC;
    avs_control_read         : IN  STD_LOGIC;
    avs_control_writedata    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- use default MM bus width for control
    avs_control_readdata     : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    -- MM slave I2C protocol register
    avs_protocol_address     : IN  STD_LOGIC_VECTOR(g_protocol_adr_w-1 DOWNTO 0);
    avs_protocol_write       : IN  STD_LOGIC;
    avs_protocol_read        : IN  STD_LOGIC;
    avs_protocol_writedata   : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    avs_protocol_readdata    : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    -- MM slave I2C result register
    avs_result_address       : IN  STD_LOGIC_VECTOR(g_result_adr_w-1 DOWNTO 0);
    avs_result_write         : IN  STD_LOGIC;
    avs_result_read          : IN  STD_LOGIC;
    avs_result_writedata     : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    avs_result_readdata      : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    
    ---------------------------------------------------------------------------
    -- Avalon Interrupt Sender interface: ins_*
    ---------------------------------------------------------------------------
    ins_interrupt_irq        : OUT STD_LOGIC
  );
END avs_i2c_master;


ARCHITECTURE wrap OF avs_i2c_master IS

  CONSTANT c_avs_i2c_mm  : t_c_i2c_mm  := (g_control_adr_w, g_protocol_adr_w, 2**g_protocol_adr_w, g_result_adr_w, 2**g_result_adr_w);
  CONSTANT c_avs_i2c_phy : t_c_i2c_phy := (g_clk_cnt, g_comma_w);

  
  SIGNAL cs_sim                : BOOLEAN;
  SIGNAL i_avs_control_address : STD_LOGIC_VECTOR(0 DOWNTO 0);

BEGIN

  -- SOPC Builder does not use BOOLEAN, so use STD_LOGIC and is_true()
  cs_sim <= is_true(coe_gs_sim_export);

  -- SOPC Builder writes STD_LOGIC_VECTOR(0 DOWNTO 0) as STD_LOGIC so take care of mapping
  i_avs_control_address(0) <= avs_control_address;
  
  u_i2c_master : ENTITY work.i2c_master
  GENERIC MAP (
    g_i2c_mm                 => c_avs_i2c_mm,
    g_i2c_phy                => c_avs_i2c_phy
  )
  PORT MAP (
    -- GENERIC Signal
    gs_sim                   => cs_sim,
    
    rst                      => csi_system_reset,
    clk                      => csi_system_clk,
    sync                     => coe_sync_export,
    
    ---------------------------------------------------------------------------
    -- Memory Mapped Slave interface with Interrupt
    ---------------------------------------------------------------------------
    -- MM slave I2C control register
    mms_control_address      => i_avs_control_address,
    mms_control_write        => avs_control_write,
    mms_control_read         => avs_control_read,
    mms_control_writedata    => avs_control_writedata,
    mms_control_readdata     => avs_control_readdata,
    -- MM slave I2C protocol register
    mms_protocol_address     => avs_protocol_address,
    mms_protocol_write       => avs_protocol_write,
    mms_protocol_read        => avs_protocol_read,
    mms_protocol_writedata   => avs_protocol_writedata,
    mms_protocol_readdata    => avs_protocol_readdata,
    -- MM slave I2C result register
    mms_result_address       => avs_result_address,
    mms_result_write         => avs_result_write,
    mms_result_read          => avs_result_read,
    mms_result_writedata     => avs_result_writedata,
    mms_result_readdata      => avs_result_readdata,
    -- Interrupt
    ins_result_rdy           => ins_interrupt_irq,
    
    ---------------------------------------------------------------------------
    -- I2C interface
    ---------------------------------------------------------------------------
    scl                      => coe_i2c_scl_export,
    sda	                     => coe_i2c_sda_export
  );
  
END wrap;

