-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: I2C commander state machine
-- Description: See i2c_commander.vhd
-- Remark:

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.i2c_pkg.ALL;
USE work.i2c_commander_pkg.ALL;

ENTITY i2c_commander_ctrl IS
  GENERIC (
    g_i2c_cmdr  : t_c_i2c_cmdr_commander;
    g_i2c_mm    : t_c_i2c_mm := c_i2c_mm
  );
  PORT (
    -- Clocks and reset
    rst                      : IN  STD_LOGIC;  -- reset synchronous with clk
    clk                      : IN  STD_LOGIC;  -- memory-mapped bus clock
    
    -- MM registers
    protocol_wr              : IN  STD_LOGIC;
    protocol_index           : IN  NATURAL RANGE 0 TO g_i2c_cmdr.nof_protocols-1;
    protocol_offset_arr      : IN  t_natural_arr(0 TO g_i2c_cmdr.nof_protocols-1);
    result_expected_arr      : IN  t_slv_32_arr(0 TO g_i2c_cmdr.nof_protocols-1);
    protocol_status_rd       : IN  STD_LOGIC;
    protocol_status          : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    result_error_cnt         : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    result_data_arr          : OUT t_slv_32_arr(0 TO g_i2c_cmdr.nof_result_data_max-1);
    
    -- I2C protocol control
    protocol_offset          : OUT NATURAL;
    protocol_activate        : OUT STD_LOGIC;
    protocol_activate_ack    : IN  STD_LOGIC;
    result_wr_en             : IN  STD_LOGIC;
    result_wr_adr            : IN  STD_LOGIC_VECTOR(g_i2c_mm.result_adr_w-1 DOWNTO 0);
    result_wr_dat            : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    result_ready_evt         : IN  STD_LOGIC
  );
END ENTITY;


ARCHITECTURE rtl OF i2c_commander_ctrl IS

  CONSTANT c_expected_bi_w       : NATURAL := ceil_log2(c_word_w);
  
  SIGNAL state                   : NATURAL RANGE 0 TO c_i2c_cmdr_state_max;
  SIGNAL nxt_state               : NATURAL;
  SIGNAL i_protocol_activate     : STD_LOGIC;
  SIGNAL nxt_protocol_activate   : STD_LOGIC;
  SIGNAL i_protocol_offset       : NATURAL;
  SIGNAL nxt_protocol_offset     : NATURAL;
  SIGNAL result_expected         : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL nxt_result_expected     : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL expected_bi             : NATURAL RANGE 0 TO c_word_w-1;   -- bit index in result_expected
  SIGNAL error_cnt               : STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
  SIGNAL nxt_error_cnt           : STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
  SIGNAL data_cnt                : NATURAL RANGE 0 TO g_i2c_cmdr.nof_result_data_max;
  SIGNAL nxt_data_cnt            : NATURAL;
  SIGNAL data_arr                : t_slv_32_arr(0 TO g_i2c_cmdr.nof_result_data_max);
  SIGNAL nxt_data_arr            : t_slv_32_arr(0 TO g_i2c_cmdr.nof_result_data_max);

   
BEGIN
  
  protocol_offset   <= i_protocol_offset;
  protocol_activate <= i_protocol_activate;
  
  protocol_status  <= TO_UVEC(state, c_word_w);
  result_error_cnt <= RESIZE_UVEC(error_cnt, c_word_w);
  result_data_arr  <= data_arr(0 TO g_i2c_cmdr.nof_result_data_max-1);
  
  p_clk : PROCESS(rst, clk)
  BEGIN
    IF rst='1' THEN
      state               <= c_i2c_cmdr_state_idle;
      i_protocol_activate <= '0';
      i_protocol_offset   <= 0;
      result_expected     <= (OTHERS=>'0');
      error_cnt           <= (OTHERS=>'0');
      data_cnt            <= 0;
      data_arr            <= (OTHERS=>(OTHERS=>'0'));
    ELSIF rising_edge(clk) THEN
      state               <= nxt_state;
      i_protocol_activate <= nxt_protocol_activate;
      i_protocol_offset   <= nxt_protocol_offset;
      result_expected     <= nxt_result_expected;
      error_cnt           <= nxt_error_cnt;
      data_cnt            <= nxt_data_cnt;
      data_arr            <= nxt_data_arr;
    END IF;
  END PROCESS;
  
  p_state : PROCESS(state, i_protocol_offset, result_expected,
                    protocol_index, protocol_offset_arr, result_expected_arr,
                    protocol_wr, protocol_activate_ack, result_ready_evt, protocol_status_rd)
  BEGIN
    nxt_state <= state;
    nxt_protocol_activate <= '0';
    nxt_protocol_offset <= i_protocol_offset;
    nxt_result_expected <= result_expected;
    CASE state IS
      WHEN c_i2c_cmdr_state_idle =>
        IF protocol_wr='1' THEN
          nxt_state <= c_i2c_cmdr_state_pending;
          nxt_protocol_activate <= '1';
          nxt_protocol_offset <= protocol_offset_arr(protocol_index);
          nxt_result_expected <= result_expected_arr(protocol_index);
        END IF;
      WHEN c_i2c_cmdr_state_pending =>
        IF protocol_activate_ack='1' THEN
          nxt_state <= c_i2c_cmdr_state_busy;
        END IF;
      WHEN c_i2c_cmdr_state_busy =>
        IF result_ready_evt='1' THEN
          nxt_state <= c_i2c_cmdr_state_done;
        END IF;
      WHEN OTHERS =>  -- = c_i2c_cmdr_state_done
        IF protocol_status_rd='1' THEN
          nxt_state <= c_i2c_cmdr_state_idle;
        END IF;
    END CASE;
  END PROCESS;
  
  -- The bits in result_expected mark the control and data bytes, it is used periodically, so for more than 32 result bytes the expected result must be the same
  -- . for protocol lists with only control and write I2C accesses all result_expected mask bits are '0'
  -- . for protocol lists with both control, write I2C and read I2C accesses only the reads have result_expected mask bit '1'
  expected_bi <= TO_UINT(result_wr_adr(c_expected_bi_w-1 DOWNTO 0));
  
  p_result : PROCESS(error_cnt, data_cnt, data_arr, i_protocol_activate, result_wr_en, result_expected, expected_bi, result_wr_dat)
  BEGIN
    nxt_error_cnt <= error_cnt;
    nxt_data_cnt  <= data_cnt;
    nxt_data_arr  <= data_arr;
    IF i_protocol_activate='1' THEN
      nxt_error_cnt <= (OTHERS=>'0');
      nxt_data_cnt  <= 0;
      nxt_data_arr  <= (OTHERS=>(OTHERS=>'0'));
    ELSIF result_wr_en='1' THEN
      IF result_expected(expected_bi)='0' THEN
        -- the result_expected at the bit index contains '0' so it marks a read control byte that has to be zero to be OK
        IF TO_UINT(result_wr_dat)/=0 THEN
          nxt_error_cnt <= INCR_UVEC(error_cnt, 1);
        END IF;
      ELSE
        -- the result_expected at the bit index contains '1' so it marks an I2C read data byte
        nxt_data_arr(data_cnt) <= RESIZE_UVEC(result_wr_dat, c_word_w);  -- store one I2C read data byte per MM word
        nxt_data_cnt <= data_cnt + 1;
      END IF;
    END IF;
  END PROCESS;
  
END rtl;
