-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE work.i2c_pkg.ALL;
USE work.i2c_smbus_pkg.ALL;

PACKAGE i2c_commander_pkg IS

  -- I2C commander settings
  --
  CONSTANT c_i2c_cmdr_state_idle    : NATURAL := 0;  -- no protocol active
  CONSTANT c_i2c_cmdr_state_pending : NATURAL := 1;  -- protocol pending by write event, but not yet activated by sync
  CONSTANT c_i2c_cmdr_state_busy    : NATURAL := 2;  -- protocol I2C accesses are busy
  CONSTANT c_i2c_cmdr_state_done    : NATURAL := 3;  -- protocol I2C accesses finished
  CONSTANT c_i2c_cmdr_state_max     : NATURAL := 3;
  CONSTANT c_i2c_cmdr_state_w       : NATURAL := ceil_log2(c_i2c_cmdr_state_max);
                                                             
  CONSTANT c_i2c_cmdr_max_nof_protocols : NATURAL := 16;  -- fixed maximum number of protocol lists that t_c_i2cmdr record can support
  CONSTANT c_i2c_cmdr_result_cnt_w      : NATURAL := 16;  -- more than sufficiently large range for most practical cases
  CONSTANT c_i2c_cmdr_nof_read_byte_max : NATURAL := 8;   -- Assume 8 read bytes per protocol list are sufficent, so c_i2c_cmdr_nof_read_byte_max=8 and use t_i2c_cmdr_natural_arr with range 0 to 7
  
  TYPE t_i2c_cmdr_natural_arr IS ARRAY (INTEGER RANGE 0 TO c_i2c_cmdr_nof_read_byte_max-1) OF NATURAL;
  TYPE t_i2c_cmdr_natural_mat IS ARRAY (INTEGER RANGE <>) OF t_i2c_cmdr_natural_arr;

  TYPE t_c_i2c_cmdr_commander IS RECORD
    nof_protocols         : NATURAL;  -- must be <= c_i2cmdr_max_nof_protocols
    protocol_offset_arr   : t_natural_arr(0 TO c_i2c_cmdr_max_nof_protocols-1);
    result_expected_arr   : t_slv_32_arr(0 TO c_i2c_cmdr_max_nof_protocols-1);
    result_cnt_w          : NATURAL;  -- the nof result bytes that a protocol list may yield is <= 2**c_i2c_protocol_adr_w
    nof_result_data_max   : NATURAL;  -- nof data bytes that a protocol list may maximally read
  END RECORD;
  
  FUNCTION func_i2c_cmdr_mm_reg_nof_dat(rec : t_c_i2c_cmdr_commander) RETURN NATURAL;
  
  FUNCTION func_i2c_cmdr_sel_a_b(sel : BOOLEAN; a, b : t_i2c_cmdr_natural_arr) RETURN t_i2c_cmdr_natural_arr;
  FUNCTION func_i2c_cmdr_sel_a_b(sel : BOOLEAN; a, b : t_c_i2c_cmdr_commander) RETURN t_c_i2c_cmdr_commander;
    
  -- Void protocol lists for the I2C commander
  CONSTANT c_i2c_cmdr_protocol_list_nop             : t_nat_natural_arr(0 DOWNTO 0) := (OTHERS=>SMBUS_C_NOP);
  CONSTANT c_i2c_cmdr_protocol_list_end             : t_nat_natural_arr(0 DOWNTO 0) := (OTHERS=>SMBUS_C_END);
  CONSTANT c_i2c_cmdr_protocol_list_sample_sda      : t_nat_natural_arr := (SMBUS_C_SAMPLE_SDA, 0, 0, 0, 0, SMBUS_C_END);

  -- Expected result list for the protocol lists, read yields (data, ok=0), write or control (like end, timeout) yield (ok=0):
  -- . entries 0 should also be 0 in the result buffer
  -- . entries 1 indicate a read octet that needs to be stored for the user
  CONSTANT c_i2c_cmdr_expected_mask_none            : STD_LOGIC_VECTOR := RESIZE_UVEC("0", c_word_w);
  CONSTANT c_i2c_cmdr_expected_mask_end             : STD_LOGIC_VECTOR := RESIZE_UVEC("0", c_word_w);
  CONSTANT c_i2c_cmdr_expected_mask_sample_sda      : STD_LOGIC_VECTOR := RESIZE_UVEC("0", c_word_w);
  CONSTANT c_i2c_cmdr_expected_mask_read_one_byte   : STD_LOGIC_VECTOR := RESIZE_UVEC("01", c_word_w);
  CONSTANT c_i2c_cmdr_expected_mask_read_one_word   : STD_LOGIC_VECTOR := RESIZE_UVEC("011", c_word_w);

  CONSTANT c_i2c_cmdr_nof_result_data_none          : NATURAL := 0;
  CONSTANT c_i2c_cmdr_nof_result_data_end           : NATURAL := 0;
  CONSTANT c_i2c_cmdr_nof_result_data_sample_sda    : NATURAL := 0;
  CONSTANT c_i2c_cmdr_nof_result_data_read_one_byte : NATURAL := 1;
  CONSTANT c_i2c_cmdr_nof_result_data_read_one_word : NATURAL := 2;
  
  CONSTANT c_i2c_cmdr_expected_x                    : NATURAL := 254;  -- do not use 255, because that matches pull-up all ones
  CONSTANT c_i2c_cmdr_expected_data_none_arr        : t_i2c_cmdr_natural_arr := (c_i2c_cmdr_expected_x, c_i2c_cmdr_expected_x,
                                                                                 c_i2c_cmdr_expected_x, c_i2c_cmdr_expected_x,
                                                                                 c_i2c_cmdr_expected_x, c_i2c_cmdr_expected_x,
                                                                                 c_i2c_cmdr_expected_x, c_i2c_cmdr_expected_x);
                                                                                 
  CONSTANT c_i2c_cmdr_mem_block_sz                  : NATURAL := 1024;  -- assign t_c_mem.nof_dat in blocks of 1024 = 1 M9K
  
END i2c_commander_pkg;


PACKAGE BODY i2c_commander_pkg IS

  FUNCTION func_i2c_cmdr_mm_reg_nof_dat(rec : t_c_i2c_cmdr_commander) RETURN NATURAL IS
  -- Example MM register map defined for:
  -- . c_nof_protocols   = 2
  -- . c_nof_result_data_max = 3
  -- 
  --    0 = protocol 0
  --    1 = protocol 1
  --    2 = protocol offset 0
  --    3 = protocol offset 1
  --    4 = result_expected 0
  --    5 = result_expected 1
  --    6 = protocol status
  --    7 = result_error_cnt
  --    8 = result_data byte 0
  --    9 = result_data byte 1
  --   10 = result_data byte 2
  -- 
  -- so return mm_reg_nof_dat = 3*2 + 1 + 1 + 3 = 11
  BEGIN
    RETURN 3*rec.nof_protocols + 1 + 1 + rec.nof_result_data_max;
  END;
  
  FUNCTION func_i2c_cmdr_sel_a_b(sel : BOOLEAN; a, b : t_i2c_cmdr_natural_arr) RETURN t_i2c_cmdr_natural_arr IS
  BEGIN
    IF sel = TRUE THEN
      RETURN a;
    ELSE
      RETURN b;
    END IF;
  END;
  
  FUNCTION func_i2c_cmdr_sel_a_b(sel : BOOLEAN; a, b : t_c_i2c_cmdr_commander) RETURN t_c_i2c_cmdr_commander IS
  BEGIN
    IF sel = TRUE THEN
      RETURN a;
    ELSE
      RETURN b;
    END IF;
  END;
  
END i2c_commander_pkg;
