-------------------------------------------------------------------------------
--
-- Copyright (C) 2011
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Provide MM slave register for I2C commander
-- Description: See i2c-commander.vhd
-- Remark:
-- . Warning "Case choice must be a locally static expression" due to using
--   operations on constants in CASE statement, can be avoid if necessary by
--   using elsif 

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.i2c_pkg.ALL;
USE work.i2c_commander_pkg.ALL;

ENTITY i2c_commander_reg IS
  GENERIC (
    g_i2c_cmdr  : t_c_i2c_cmdr_commander;
    g_i2c_mm    : t_c_i2c_mm := c_i2c_mm
  );
  PORT (
    -- Clocks and reset
    rst                      : IN  STD_LOGIC;  -- reset synchronous with clk
    clk                      : IN  STD_LOGIC;  -- memory-mapped bus clock
    
    -- Memory Mapped Slave
    sla_in                   : IN  t_mem_mosi;  -- actual ranges defined by c_mm_reg
    sla_out                  : OUT t_mem_miso;  -- actual ranges defined by c_mm_reg
    
    -- MM registers
    protocol_wr              : OUT STD_LOGIC;
    protocol_index           : OUT NATURAL RANGE 0 TO g_i2c_cmdr.nof_protocols-1;
    protocol_offset_arr      : OUT t_natural_arr(0 TO g_i2c_cmdr.nof_protocols-1);
    result_expected_arr      : OUT t_slv_32_arr(0 TO g_i2c_cmdr.nof_protocols-1);
    protocol_status_rd       : OUT STD_LOGIC;
    protocol_status          : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    result_error_cnt         : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    result_data_arr          : IN  t_slv_32_arr(0 TO g_i2c_cmdr.nof_result_data_max-1)
  );
END i2c_commander_reg;


ARCHITECTURE rtl OF i2c_commander_reg IS

  -- Use shorter aliases for some generic constants
  CONSTANT c_nof_protocols       : NATURAL := g_i2c_cmdr.nof_protocols;    -- must be <= c_i2c_max_nof_protocols to fit the t_c_i2c_cmdr_commander record
  CONSTANT c_nof_result_data_max : NATURAL := g_i2c_cmdr.nof_result_data_max;
  
  CONSTANT c_protocol_adr_w      : NATURAL := g_i2c_mm.protocol_adr_w;
  
  -- Example MM register map defined for:
  -- . c_nof_protocols   = 2
  -- . c_nof_result_data_max = 3
  -- 
  --    0 = protocol 0
  --    1 = protocol 1
  --    2 = protocol offset 0
  --    3 = protocol offset 1
  --    4 = result_expected 0
  --    5 = result_expected 1
  --    6 = protocol status
  --    7 = result_error_cnt
  --    8 = result_data byte 0
  --    9 = result_data byte 1
  --   10 = result_data byte 2
  -- 
  -- so c_mm_reg_nof_dat = 3*2 + 1 + 1 + 3 = 11
  
  CONSTANT c_mm_reg_nof_dat     : NATURAL := func_i2c_cmdr_mm_reg_nof_dat(g_i2c_cmdr);
  CONSTANT c_mm_reg_w           : NATURAL := ceil_log2(c_mm_reg_nof_dat);
  
  CONSTANT c_mm_reg             : t_c_mem := (latency  => 1,
                                              adr_w    => c_mm_reg_w,
                                              dat_w    => c_word_w,       -- Use MM bus data width = c_word_w = 32 for all MM registers
                                              nof_dat  => c_mm_reg_nof_dat,
                                              init_sl  => '0');

  -- Commander registers
  SIGNAL i_protocol_offset_arr  : t_natural_arr(0 TO c_nof_protocols-1);
  SIGNAL i_result_expected_arr  : t_slv_32_arr(0 TO c_nof_protocols-1);
  
BEGIN

  protocol_offset_arr <= i_protocol_offset_arr;
  result_expected_arr <= i_result_expected_arr;
  
  ------------------------------------------------------------------------------
  -- MM register access in the MM clk domain
  -- . Hardcode the shared MM slave register directly in RTL instead of using
  --   the common_reg_r_w instance. Directly using RTL is easier when the large
  --   MM register has multiple different fields and with different read and
  --   write options per field in one MM register.
  ------------------------------------------------------------------------------

  p_mm_reg : PROCESS (rst, clk)
    VARIABLE v_address           : NATURAL := 0;
    VARIABLE v_index_protocol    : NATURAL := 0;
    VARIABLE v_index_offset      : INTEGER := 0;
    VARIABLE v_index_expected    : INTEGER := 0;
    VARIABLE v_index_result_data : INTEGER := 0;
  BEGIN
    IF rst = '1' THEN
    
      -- Read access
      sla_out <= c_mem_miso_rst;
      
      -- Access event, register values
      protocol_wr <= '0';
      
      i_protocol_offset_arr <= g_i2c_cmdr.protocol_offset_arr(0 TO c_nof_protocols-1);
      i_result_expected_arr <= g_i2c_cmdr.result_expected_arr(0 TO c_nof_protocols-1);
      
    ELSIF rising_edge(clk) THEN
      -- Init auxiliary semi-constant variables (do it here within rising_edge section to avoid having to put sla_in in sensitivity list)
      v_address           := TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0));
      v_index_protocol    := v_address;
      v_index_offset      := v_address -   c_nof_protocols;
      v_index_expected    := v_address - 2*c_nof_protocols;
      v_index_result_data := v_address - 3*c_nof_protocols - 2;
   
      -- Read access defaults
      sla_out.rdval <= '0';
      
      -- Access event defaults
      protocol_wr <= '0';
      
      -- Write access: set register value
      IF sla_in.wr = '1' THEN
        CASE v_address IS
        
          -- Write commander access indicates the new protocol event, ignore the wrdata
          WHEN 0 TO c_nof_protocols-1 =>
            protocol_wr <= '1';
            protocol_index <= v_index_protocol;
          
          -- Write offset address per commander protocol
          WHEN c_nof_protocols TO 2*c_nof_protocols-1 => i_protocol_offset_arr(v_index_offset) <= TO_UINT(sla_in.wrdata(c_protocol_adr_w-1 DOWNTO 0));
            
          -- Write expected periodic result pattern per commander protocol
          WHEN 2*c_nof_protocols TO 3*c_nof_protocols-1 => i_result_expected_arr(v_index_expected) <= sla_in.wrdata(31 DOWNTO 0);
            
          -- The protocol status is read only
          -- The nof result bytes count is read only
          -- The result data bytes are read only
          -- Not used MM addresses
          WHEN OTHERS => NULL;
        END CASE;
        
      -- Read access: get register value
      ELSIF sla_in.rd = '1' THEN
        sla_out       <= c_mem_miso_rst;    -- set unused rddata bits to '0' when read
        sla_out.rdval <= '1';               -- c_mm_reg.latency = 1
        CASE v_address IS
        
          -- Read offset address per commander protocol
          WHEN c_nof_protocols TO 2*c_nof_protocols-1 => sla_out.rddata(31 DOWNTO 0) <= TO_UVEC(i_protocol_offset_arr(v_index_offset), c_word_w);
            
          -- Read expected periodic result pattern per commander protocol
          WHEN 2*c_nof_protocols TO 3*c_nof_protocols-1 => sla_out.rddata(31 DOWNTO 0) <= i_result_expected_arr(v_index_expected);
            
          -- Read the commander status
          WHEN 3*c_nof_protocols => sla_out.rddata(31 DOWNTO 0) <= protocol_status;
          
          -- Read the nof result errors count
          -- The nof result errors count is 0 when OK, else it shows how often the result byte differed from 0 as indicated by '0' in the expected result pattern
          WHEN 3*c_nof_protocols+1 => sla_out.rddata(31 DOWNTO 0) <= result_error_cnt;
          
          -- The result data has 1 byte per address and is read only, but do support overwriting the register
          -- The result data are the result bytes as indicated by '1' in the expected result pattern
          WHEN 3*c_nof_protocols+2 TO 3*c_nof_protocols+2+c_nof_result_data_max-1 => sla_out.rddata(31 DOWNTO 0) <= RESIZE_UVEC(result_data_arr(v_index_result_data), c_word_w);
            
          -- The commander triggers are write access only        
          -- Not used MM addresses
          WHEN OTHERS => NULL;
        END CASE;
      END IF;
    END IF;
  END PROCESS;

  -- Detect read commander status event (has to be combinatorial to avoid missing the c_i2c_state_done state, because the state depends on the this pulse)
  protocol_status_rd <= '1' WHEN sla_in.rd = '1' AND TO_UINT(sla_in.address(c_mm_reg.adr_w-1 DOWNTO 0))=3*c_nof_protocols ELSE '0';
  
END rtl;
