-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;

ENTITY i2c_list_ctrl IS
  GENERIC (
    g_protocol_adr_w : NATURAL := 10;
    g_result_adr_w   : NATURAL := 10
  );
  PORT (
    clk              : IN  STD_LOGIC;
    rst              : IN  STD_LOGIC;
    activate         : IN  STD_LOGIC;
    busy             : OUT STD_LOGIC;
    protocol_rd_en   : OUT STD_LOGIC;
    protocol_rd_adr  : OUT STD_LOGIC_VECTOR(g_protocol_adr_w-1 DOWNTO 0);
    protocol_rd_ofs  : IN  NATURAL := 0;
    protocol_rd_dat  : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    protocol_rd_val  : IN  STD_LOGIC;
    result_wr_en     : OUT STD_LOGIC;
    result_wr_adr    : OUT STD_LOGIC_VECTOR(g_result_adr_w-1 DOWNTO 0);
    result_wr_dat    : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    smbus_out_dat    : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    smbus_out_req    : OUT STD_LOGIC;    
    smbus_in_dat     : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    smbus_in_val     : IN  STD_LOGIC;
    smbus_in_err     : IN  STD_LOGIC; 
    smbus_in_ack     : IN  STD_LOGIC;
    smbus_st_idle    : IN  STD_LOGIC;
    smbus_st_end     : IN  STD_LOGIC
  );
END ENTITY;


ARCHITECTURE rtl OF i2c_list_ctrl IS

  SIGNAL i_busy               : STD_LOGIC;
  SIGNAL nxt_busy             : STD_LOGIC;
  SIGNAL nxt_protocol_rd_en   : STD_LOGIC;
  SIGNAL i_protocol_rd_adr    : STD_LOGIC_VECTOR(g_protocol_adr_w-1 DOWNTO 0);
  SIGNAL nxt_protocol_rd_adr  : STD_LOGIC_VECTOR(g_protocol_adr_w-1 DOWNTO 0);
  SIGNAL i_result_wr_en       : STD_LOGIC;
  SIGNAL nxt_result_wr_en     : STD_LOGIC;
  SIGNAL i_result_wr_adr      : STD_LOGIC_VECTOR(g_result_adr_w-1 DOWNTO 0);
  SIGNAL nxt_result_wr_adr    : STD_LOGIC_VECTOR(g_result_adr_w-1 DOWNTO 0);
  SIGNAL nxt_result_wr_dat    : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  SIGNAL i_smbus_in_err       : STD_LOGIC_VECTOR(0 DOWNTO 0);
  SIGNAL prev_smbus_st_idle   : STD_LOGIC;
  SIGNAL prot_done            : STD_LOGIC;
  SIGNAL pend_st_end          : STD_LOGIC;
  SIGNAL nxt_pend_st_end      : STD_LOGIC;
  SIGNAL list_end             : STD_LOGIC;
  SIGNAL prev_list_end        : STD_LOGIC;
  SIGNAL i_smbus_out_req      : STD_LOGIC;
  SIGNAL nxt_smbus_out_req    : STD_LOGIC;
  
BEGIN

  protocol_rd_adr   <= i_protocol_rd_adr;
  
  result_wr_en      <= i_result_wr_en;
  result_wr_adr     <= i_result_wr_adr;
  
  i_smbus_in_err(0) <= smbus_in_err;
  
  smbus_out_req     <= i_smbus_out_req;
  smbus_out_dat     <= protocol_rd_dat;
  
  busy              <= i_busy;
  prot_done         <= NOT prev_smbus_st_idle AND smbus_st_idle;
  list_end          <= smbus_st_end OR pend_st_end;
  
  regs: PROCESS(rst,clk)
  BEGIN
    IF rst='1' THEN
      i_busy             <= '0';
      pend_st_end        <= '0';
      protocol_rd_en     <= '0';
      i_protocol_rd_adr  <= (OTHERS => '0');
      i_result_wr_en     <= '0';
      i_result_wr_adr    <= (OTHERS => '0');
      result_wr_dat      <= (OTHERS => '0');
      prev_smbus_st_idle <= '1';
      prev_list_end      <= '0';
      i_smbus_out_req    <= '0';
    ELSIF RISING_EDGE(clk) THEN
      i_busy             <= nxt_busy;
      pend_st_end        <= nxt_pend_st_end;
      protocol_rd_en     <= nxt_protocol_rd_en;
      i_protocol_rd_adr  <= nxt_protocol_rd_adr;
      i_result_wr_en     <= nxt_result_wr_en;
      i_result_wr_adr    <= nxt_result_wr_adr;
      result_wr_dat      <= nxt_result_wr_dat;
      prev_smbus_st_idle <= smbus_st_idle;
      prev_list_end      <= list_end;
      i_smbus_out_req    <= nxt_smbus_out_req;
    END IF;
  END PROCESS;
  
  do_protocol_list : PROCESS(activate, i_busy, prev_list_end, list_end, smbus_in_ack, i_protocol_rd_adr, protocol_rd_ofs)
  BEGIN
    nxt_busy <= i_busy;
    IF activate='1' THEN
      nxt_busy <= '1';
    ELSIF prev_list_end='1' AND list_end='0' THEN
      nxt_busy <= '0';
    END IF;

    nxt_protocol_rd_en  <= '0';
    nxt_protocol_rd_adr <= i_protocol_rd_adr;
    IF activate='1' THEN
      nxt_protocol_rd_en  <= '1';
      nxt_protocol_rd_adr <= TO_UVEC(protocol_rd_ofs, g_protocol_adr_w);
    ELSIF smbus_in_ack='1' AND list_end='0' THEN
      nxt_protocol_rd_en  <= '1';
      nxt_protocol_rd_adr <= STD_LOGIC_VECTOR(UNSIGNED(i_protocol_rd_adr)+1);
    END IF;
  END PROCESS;
  
  smbus_request : PROCESS(i_smbus_out_req, protocol_rd_val, list_end, smbus_in_ack)
  BEGIN
    nxt_smbus_out_req <= i_smbus_out_req;
    IF protocol_rd_val='1' AND list_end='0' THEN
      nxt_smbus_out_req <= '1';
    ELSIF smbus_in_ack='1' THEN
      nxt_smbus_out_req <= '0';
    END IF;
  END PROCESS;

  end_of_protocol_list: PROCESS(pend_st_end, prot_done, smbus_st_end)
  BEGIN
    nxt_pend_st_end <= pend_st_end;
    IF prot_done='1' THEN
      nxt_pend_st_end <= '0';
    ELSIF smbus_st_end='1' THEN
      nxt_pend_st_end <= '1';
    END IF;
  END PROCESS;
  
  save_results : PROCESS(activate, i_result_wr_adr, i_result_wr_en, prot_done, smbus_in_val, i_smbus_in_err, smbus_in_dat)
  BEGIN    
    nxt_result_wr_adr <= i_result_wr_adr;
    IF activate='1' THEN
      nxt_result_wr_adr  <= (OTHERS=>'0');
    ELSIF i_result_wr_en='1' THEN
      nxt_result_wr_adr <= STD_LOGIC_VECTOR(UNSIGNED(i_result_wr_adr)+1);
    END IF;
    
    nxt_result_wr_en  <= '0';
    nxt_result_wr_dat <= (OTHERS => '0');
    IF prot_done='1' THEN
      nxt_result_wr_en   <= '1';
      nxt_result_wr_dat  <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(i_smbus_in_err), c_byte_w));
    ELSIF smbus_in_val='1' THEN
      nxt_result_wr_en   <= '1';
      nxt_result_wr_dat  <= smbus_in_dat;
    END IF;
  END PROCESS;
  
END rtl;
