-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE work.i2c_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY i2c_master IS
  GENERIC (
    g_technology             : NATURAL := c_tech_select_default;
    g_i2c_mm                 : t_c_i2c_mm  := c_i2c_mm;
    g_i2c_phy                : t_c_i2c_phy
  );
  PORT (
    -- GENERIC Signal
    gs_sim                   : IN  BOOLEAN := FALSE;
    
    rst                      : IN  STD_LOGIC;
    clk                      : IN  STD_LOGIC;
    sync                     : IN  STD_LOGIC := '1';
    
    ---------------------------------------------------------------------------
    -- Memory Mapped Slave interface with Interrupt
    ---------------------------------------------------------------------------
    -- MM slave I2C control register
    mms_control_address      : IN  STD_LOGIC_VECTOR(g_i2c_mm.control_adr_w-1 DOWNTO 0);
    mms_control_write        : IN  STD_LOGIC;
    mms_control_read         : IN  STD_LOGIC;
    mms_control_writedata    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- use default MM bus width for control
    mms_control_readdata     : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    -- MM slave I2C protocol register
    mms_protocol_address     : IN  STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
    mms_protocol_write       : IN  STD_LOGIC;
    mms_protocol_read        : IN  STD_LOGIC;
    mms_protocol_writedata   : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    mms_protocol_readdata    : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    -- MM slave I2C result register
    mms_result_address       : IN  STD_LOGIC_VECTOR(g_i2c_mm.result_adr_w-1 DOWNTO 0);
    mms_result_write         : IN  STD_LOGIC;
    mms_result_read          : IN  STD_LOGIC;
    mms_result_writedata     : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    mms_result_readdata      : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    -- Interrupt
    ins_result_rdy           : OUT STD_LOGIC; 
    
    ---------------------------------------------------------------------------
    -- I2C interface
    ---------------------------------------------------------------------------
    scl                      : INOUT STD_LOGIC;  -- I2C Serial Clock Line
    sda	                     : INOUT STD_LOGIC   -- I2C Serial Data Line
  );
END i2c_master;

ARCHITECTURE str OF i2c_master IS

  -- MM - I2C ctrl list
  SIGNAL protocol_rd_en         : STD_LOGIC;
  SIGNAL protocol_rd_adr        : STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
  SIGNAL protocol_rd_dat        : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  SIGNAL protocol_rd_val        : STD_LOGIC;

  SIGNAL result_wr_en           : STD_LOGIC;
  SIGNAL result_wr_adr          : STD_LOGIC_VECTOR(g_i2c_mm.result_adr_w-1 DOWNTO 0);
  SIGNAL result_wr_dat          : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  
  SIGNAL protocol_activate_evt  : STD_LOGIC;
  SIGNAL result_ready_evt       : STD_LOGIC;

  -- I2C ctrl list - SMBus
  SIGNAL smbus_in_dat           : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  SIGNAL smbus_in_req           : STD_LOGIC;    
  SIGNAL smbus_out_dat          : STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
  SIGNAL smbus_out_val          : STD_LOGIC;
  SIGNAL smbus_out_err          : STD_LOGIC;
  SIGNAL smbus_out_ack          : STD_LOGIC;
  SIGNAL smbus_st_idle          : STD_LOGIC;
  SIGNAL smbus_st_end           : STD_LOGIC;
  
BEGIN

  result_ready_evt <= smbus_st_end;

  u_mm : ENTITY work.i2c_mm
  GENERIC MAP (
    g_technology => g_technology,
    g_i2c_mm => g_i2c_mm
  )
  PORT MAP (
    rst                      => rst,
    clk                      => clk,
    sync                     => sync,
    ---------------------------------------------------------------------------
    -- Memory Mapped Slave interface with Interrupt
    ---------------------------------------------------------------------------
    -- MM slave I2C control register
    mms_control_address      => mms_control_address,
    mms_control_write        => mms_control_write,
    mms_control_read         => mms_control_read,
    mms_control_writedata    => mms_control_writedata,
    mms_control_readdata     => mms_control_readdata,
    -- MM slave I2C protocol register
    mms_protocol_address     => mms_protocol_address,
    mms_protocol_write       => mms_protocol_write,
    mms_protocol_read        => mms_protocol_read,
    mms_protocol_writedata   => mms_protocol_writedata,
    mms_protocol_readdata    => mms_protocol_readdata,
    -- MM slave I2C result register
    mms_result_address       => mms_result_address,
    mms_result_write         => mms_result_write,
    mms_result_read          => mms_result_read,
    mms_result_writedata     => mms_result_writedata,
    mms_result_readdata      => mms_result_readdata,
    -- Interrupt
    ins_result_rdy           => ins_result_rdy,
    ---------------------------------------------------------------------------
    -- SMBus control interface
    ---------------------------------------------------------------------------
    protocol_rd_en           => protocol_rd_en,
    protocol_rd_adr          => protocol_rd_adr,
    protocol_rd_dat          => protocol_rd_dat,
    protocol_rd_val          => protocol_rd_val,
    
    result_wr_en             => result_wr_en,
    result_wr_adr            => result_wr_adr,
    result_wr_dat            => result_wr_dat,
    
    protocol_activate_evt    => protocol_activate_evt,
    result_ready_evt         => result_ready_evt
  );

  u_ctrl : ENTITY work.i2c_list_ctrl
  GENERIC MAP (
    g_protocol_adr_w => g_i2c_mm.protocol_adr_w,
    g_result_adr_w   => g_i2c_mm.result_adr_w
  )
  PORT MAP (
    rst              => rst,
    clk              => clk,
    activate         => protocol_activate_evt,
    busy             => OPEN,
    protocol_rd_en   => protocol_rd_en,
    protocol_rd_adr  => protocol_rd_adr,
    protocol_rd_dat  => protocol_rd_dat,
    protocol_rd_val  => protocol_rd_val,
    result_wr_en     => result_wr_en,
    result_wr_adr    => result_wr_adr,
    result_wr_dat    => result_wr_dat,
    smbus_out_dat    => smbus_in_dat,
    smbus_out_req    => smbus_in_req,
    smbus_in_dat     => smbus_out_dat,
    smbus_in_val     => smbus_out_val,
    smbus_in_err     => smbus_out_err, 
    smbus_in_ack     => smbus_out_ack,
    smbus_st_idle    => smbus_st_idle,
    smbus_st_end     => smbus_st_end 
  );

  u_smbus : ENTITY work.i2c_smbus
  GENERIC MAP (
    g_i2c_phy        => g_i2c_phy
  )
  PORT MAP (
    gs_sim           => gs_sim,
    rst              => rst,
    clk              => clk,
    in_dat           => smbus_in_dat,
    in_req           => smbus_in_req,
    out_dat          => smbus_out_dat,
    out_val          => smbus_out_val,
    out_err          => smbus_out_err,
    out_ack          => smbus_out_ack,
    st_idle          => smbus_st_idle,
    st_end           => smbus_st_end,
    scl              => scl,
    sda              => sda
  );

END str;
