-------------------------------------------------------------------------------
--
-- Copyright (C) 2009
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, technology_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE work.i2c_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;

ENTITY i2c_mm IS
  GENERIC (
    g_technology             : NATURAL := c_tech_select_default;
    g_i2c_mm                 : t_c_i2c_mm  := c_i2c_mm
  );
  PORT (
    rst                      : IN  STD_LOGIC;
    clk                      : IN  STD_LOGIC;
    sync                     : IN  STD_LOGIC := '1';
    
    ---------------------------------------------------------------------------
    -- Avalon Memory Mapped Slave interface with Interrupt
    ---------------------------------------------------------------------------
    -- MM slave I2C control register
    mms_control_address      : IN  STD_LOGIC_VECTOR(g_i2c_mm.control_adr_w-1 DOWNTO 0);
    mms_control_write        : IN  STD_LOGIC;
    mms_control_read         : IN  STD_LOGIC;
    mms_control_writedata    : IN  STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- use default MM bus width for control
    mms_control_readdata     : OUT STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    -- MM slave I2C protocol register
    mms_protocol_address     : IN  STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
    mms_protocol_write       : IN  STD_LOGIC;
    mms_protocol_read        : IN  STD_LOGIC;
    mms_protocol_writedata   : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    mms_protocol_readdata    : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    -- MM slave I2C result register
    mms_result_address       : IN  STD_LOGIC_VECTOR(g_i2c_mm.result_adr_w-1 DOWNTO 0);
    mms_result_write         : IN  STD_LOGIC;
    mms_result_read          : IN  STD_LOGIC;
    mms_result_writedata     : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);  -- define MM bus data has same width as SMBus data
    mms_result_readdata      : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    -- Interrupt
    ins_result_rdy           : OUT STD_LOGIC; 
    
    ---------------------------------------------------------------------------
    -- SMBus control interface
    ---------------------------------------------------------------------------
    protocol_rd_en           : IN  STD_LOGIC;
    protocol_rd_adr          : IN  STD_LOGIC_VECTOR(g_i2c_mm.protocol_adr_w-1 DOWNTO 0);
    protocol_rd_dat          : OUT STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);
    protocol_rd_val          : OUT STD_LOGIC;
    
    result_wr_en             : IN  STD_LOGIC;
    result_wr_adr            : IN  STD_LOGIC_VECTOR(g_i2c_mm.result_adr_w-1 DOWNTO 0);
    result_wr_dat            : IN  STD_LOGIC_VECTOR(c_byte_w-1 DOWNTO 0);

    protocol_activate_evt    : OUT STD_LOGIC;  -- pulse indicating new protocol list
    result_ready_evt         : IN  STD_LOGIC   -- pulse indicating new result is available
  );
END i2c_mm;


ARCHITECTURE str OF i2c_mm IS

  -- Use default MM bus data width = c_word_w
  CONSTANT c_reg_control  : t_c_mem := (latency  => 1,
                                        adr_w    => 1,
                                        dat_w    => c_word_w,
                                        nof_dat  => 1,
                                        init_sl  => 'X');

  -- Use MM bus data width = SMBus data width = c_byte_w
  CONSTANT c_ram_protocol : t_c_mem := (latency  => 2,
                                        adr_w    => g_i2c_mm.protocol_adr_w,
                                        dat_w    => c_byte_w,
                                        nof_dat  => 2**g_i2c_mm.protocol_adr_w,
                                        init_sl  => 'X');
  CONSTANT c_ram_result   : t_c_mem := (latency  => 2,
                                        adr_w    => g_i2c_mm.result_adr_w,
                                        dat_w    => c_byte_w,
                                        nof_dat  => 2**g_i2c_mm.result_adr_w,
                                        init_sl  => 'X');

  -- Control register
  CONSTANT c_control_bit_activate  : NATURAL := 0;
  CONSTANT c_control_bit_ready     : NATURAL := 1;

  CONSTANT c_init_dummy            : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 2) := (OTHERS => '0');
  CONSTANT c_init_control_activate : STD_LOGIC := '0';
  CONSTANT c_init_control_ready    : STD_LOGIC := '0';
  CONSTANT c_init_control          : STD_LOGIC_VECTOR(c_mem_reg_init_w-1 DOWNTO 0) := c_init_dummy
                                                                                    & c_init_control_ready
                                                                                    & c_init_control_activate;

  SIGNAL reg_control               : STD_LOGIC_VECTOR(c_reg_control.nof_dat*c_reg_control.dat_w-1 DOWNTO 0);
  SIGNAL reg_control_rd            : STD_LOGIC_VECTOR(reg_control'RANGE);

  SIGNAL result_ready              : STD_LOGIC;

BEGIN

  -- Protocol activate control
  u_protocol_activate : ENTITY common_lib.common_request
  PORT MAP (
    rst         => rst,
    clk         => clk,
    sync        => sync,
    in_req      => reg_control(c_control_bit_activate),
    out_req_evt => protocol_activate_evt
  );

  -- Result ready control
  u_result_ready : ENTITY common_lib.common_switch
  PORT MAP (
    clk         => clk,
    rst         => rst,
    switch_high => result_ready_evt,
    switch_low  => mms_control_read,   -- read clear ready bit
    out_level   => result_ready
  );

  ins_result_rdy <= result_ready;
  
  -- I2C access control register
  p_reg_control_rd : PROCESS(reg_control, result_ready)
  BEGIN
    reg_control_rd                      <= reg_control;
    reg_control_rd(c_control_bit_ready) <= result_ready;
  END PROCESS;

  u_reg_control : ENTITY common_lib.common_reg_r_w
  GENERIC MAP (
    g_reg       => c_reg_control,
    g_init_reg  => c_init_control
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    clken       => '1',
    -- control side
    wr_en       => mms_control_write,
    wr_adr      => mms_control_address,
    wr_dat      => mms_control_writedata,
    rd_en       => mms_control_read,
    rd_adr      => mms_control_address,
    rd_dat      => mms_control_readdata,
    rd_val      => OPEN,
    -- data side
    out_reg     => reg_control,
    in_reg      => reg_control_rd
  );

  -- I2C protocol list register
  u_ram_protocol : ENTITY common_lib.common_ram_rw_rw
  GENERIC MAP (
    g_technology => g_technology,
    g_ram        => c_ram_protocol
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    clken       => '1',
    wr_en_a     => mms_protocol_write,
    wr_dat_a    => mms_protocol_writedata,
    adr_a       => mms_protocol_address,
    rd_en_a     => mms_protocol_read,
    rd_dat_a    => mms_protocol_readdata,
    rd_val_a    => OPEN,
    wr_en_b     => '0',
    --wr_dat_b    => (OTHERS => '0'),
    adr_b       => protocol_rd_adr,
    rd_en_b     => protocol_rd_en,
    rd_dat_b    => protocol_rd_dat,
    rd_val_b    => protocol_rd_val
  );

  -- I2C result register
  u_ram_result : ENTITY common_lib.common_ram_rw_rw
  GENERIC MAP (
    g_technology => g_technology,
    g_ram        => c_ram_result
  )
  PORT MAP (
    rst         => rst,
    clk         => clk,
    clken       => '1',
    wr_en_a     => mms_result_write,
    wr_dat_a    => mms_result_writedata,
    adr_a       => mms_result_address,
    rd_en_a     => mms_result_read,
    rd_dat_a    => mms_result_readdata,
    rd_val_a    => OPEN,
    wr_en_b     => result_wr_en,
    wr_dat_b    => result_wr_dat,
    adr_b       => result_wr_adr,
    rd_en_b     => '0',
    rd_dat_b    => OPEN,
    rd_val_b    => OPEN
  );

END str;
